"""Create `SPHERES` data set."""

import numpy as np

import torch

from torch.utils.data import Dataset
from torch.utils.data import random_split

from torch_topological.data import sample_from_sphere


def create_sphere_dataset(n_samples=500, n_spheres=11, d=100, r=5, seed=None):
    """Create data set of high-dimensional spheres.

    Create `SPHERES` data set described in Moor et al. [Moor20a]_. The
    data sets consists of `n` spheres, enclosed by a single sphere. It
    is a perfect example of simple manifolds, being arranged in simple
    pattern, that is nevertheless challenging to embed by algorithms.

    Parameters
    ----------
    n_samples : int
        Number of points to sample per sphere.

    n_spheres : int
        Total number of spheres to create. The algorithm will always
        create the *last* sphere to enclose the previous ones. Hence,
        if `n_spheres = 3`, two spheres will be enclosed by a larger
        one.

    d : int
        Dimension of spheres to sample from. A `d`-sphere will be
        embedded in `d+1` dimensions.

    r : float
        Radius of smaller spheres. The radius of the larger enclosing
        sphere will be `5 * r`.

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    Tuple of `np.array`, `np.array`
        Array containing the coordinates of the spheres. The second
        array contains the respective labels, ranging from `0` to
        `n_spheres - 1`. This array can be used for visualisation
        purposes.

    Notes
    -----
    The original version of this code was authored by Michael Moor.

    References
    ----------
    .. [Moor20a] M. Moor et al., "Topological Autoencoders",
        *Proceedings of the 37th International Conference on Machine
        Learning*, PMLR 119, pp. 7045--7054, 2020.
    """
    rng = np.random.default_rng(seed)

    variance = 10 / np.sqrt(d)
    shift_matrix = rng.normal(0, variance, [n_spheres, d+1])

    spheres = []
    n_datapoints = 0
    for i in np.arange(n_spheres - 1):
        sphere = sample_from_sphere(n=n_samples, d=d, r=r)
        spheres.append(sphere + shift_matrix[i, :])
        n_datapoints += n_samples

    # Build additional large surrounding sphere:
    n_samples_big = 10 * n_samples
    big = sample_from_sphere(n=n_samples_big, d=d, r=r*5)
    spheres.append(big)
    n_datapoints += n_samples_big

    X = np.concatenate(spheres, axis=0)
    y = np.zeros(n_datapoints)

    label_index = 0

    for index, data in enumerate(spheres):
        n_sphere_samples = data.shape[0]
        y[label_index:label_index + n_sphere_samples] = index
        label_index += n_sphere_samples

    return X, y


class Spheres(Dataset):
    """Data set containing multiple spheres, enclosed by a larger one.

    This is a `Dataset` instance of the `SPHERES` data set described by
    Moor et al. [Moor20a]_. The class provides a convenient interface
    for making use of that data set in machine learning tasks.
    """

    def __init__(
        self,
        train=True,
        test_fraction=0.1,
        n_samples=100,
        n_spheres=11,
        r=5,
    ):
        """Create new instance of `SPHERES` data set.

        This class wraps the `SPHERES` data set for subsequent use in
        machine learning tasks. See :func:`create_sphere_dataset` for
        more details on the supported parameters.

        Parameters
        ----------
        train : bool
            If set, create and store training portion of data set.

        test_fraction : float
            Fraction of generated samples to be used as the test portion
            of the data set.
        """
        X, y = create_sphere_dataset(
                n_samples=n_samples,
                n_spheres=n_spheres,
                r=r)

        X = torch.as_tensor(X, dtype=torch.float)

        test_size = int(test_fraction * len(X))
        train_size = len(X) - test_size

        indices = torch.as_tensor(np.arange(len(X)), dtype=torch.long)

        train_indices, test_indices = random_split(
            indices, [train_size, test_size]
        )

        X_train, X_test = X[train_indices], X[test_indices]
        y_train, y_test = y[train_indices], y[test_indices]

        self.data = X_train if train else X_test
        self.labels = y_train if train else y_test

        self.dimension = X.shape[1]

    def __getitem__(self, index):
        """Return data point and label of a specific point.

        Parameters
        ----------
        index : int
            Index of point in data set. Invalid indices will raise an
            exception.

        Returns
        -------
        tuple of `torch.tensor` and `torch.tensor`
            The point at the specific index and its label, indicating
            the sphere it belongs to. See :func:`create_sphere_dataset`
            for more details on the specifics of label assignment.
        """
        return self.data[index], self.labels[index]

    def __len__(self):
        """Return number of points in data set.

        Returns
        -------
        int
            Number of samples in data set.
        """
        return len(self.data)



"""Layers for processing persistence diagrams."""

import torch


class StructureElementLayer(torch.nn.Module):
    def __init__(
        self,
        n_elements
    ):
        super().__init__()

        self.n_elements = n_elements
        self.dim = 2    # TODO: Make configurable

        size = (self.n_elements, self.dim)

        self.centres = torch.nn.Parameter(
            torch.rand(*size)
        )

        self.sharpness = torch.nn.Parameter(
            torch.ones(*size) * 3
        )

    def forward(self, x):
        batch = torch.cat([x] * self.n_elements, 1)

        B, N, D = x.shape

        # This is a 'butchered' variant of the much nicer `SLayerExponential`
        # class by C. Hofer and R. Kwitt.
        #
        # https://c-hofer.github.io/torchph/_modules/torchph/nn/slayer.html#SLayerExponential

        centres = torch.cat([self.centres] * N, 1)
        centres = centres.view(-1, self.dim)
        centres = torch.stack([centres] * B, 0)
        centres = torch.cat((centres, 2 * batch[..., -1].unsqueeze(-1)), 2)

        sharpness = torch.pow(self.sharpness, 2)
        sharpness = torch.cat([sharpness] * N, 1)
        sharpness = sharpness.view(-1, self.dim)
        sharpness = torch.stack([sharpness] * B, 0)
        sharpness = torch.cat(
            (
                sharpness,
                torch.ones_like(batch[..., -1].unsqueeze(-1))
            ),
            2
        )

        x = centres - batch
        x = x.pow(2)
        x = torch.mul(x, sharpness)
        x = torch.nansum(x, 2)
        x = torch.exp(-x)
        x = x.view(B, self.n_elements, -1)
        x = torch.sum(x, 2)
        x = x.squeeze()

        return x



"""Module for various data operations and data set creation strategies."""

from .shapes import sample_from_annulus
from .shapes import sample_from_disk
from .shapes import sample_from_sphere
from .shapes import sample_from_torus
from .shapes import sample_from_unit_cube

__all__ = [
    'sample_from_annulus',
    'sample_from_disk',
    'sample_from_sphere',
    'sample_from_torus',
    'sample_from_unit_cube',
]



"""General utility functions.

This module contains generic utility functions that are used throughout
the code base. If a function is 'relatively small' and somewhat generic
it should be put here.
"""


def is_iterable(x):
    """Check whether variable is iterable.

    Parameters
    ----------
    x : any
        Input object.

    Returns
    -------
    bool
        `True` if `x` is iterable.
    """
    result = True

    # This is the most generic way; it also permits objects that only
    # implement the `__getitem__` interface.
    try:
        iter(x)
    except TypeError:
        result = False

    return result


def wrap_if_not_iterable(x):
    """Wrap variable in case it cannot be iterated over.

    This function provides a convenience wrapper for variables that need
    to be iterated over. If the variable is already an `iterable`, there
    is nothing to be done and it will be returned as-is. Otherwise, will
    'wrap' the variable to be the single item of a list.

    The primary purpose of this function is to make it easier for users
    to interact with certain classes: essentially, one does not have to
    think any more about single inputs vs. `iterable` inputs.

    Parameters
    ----------
    x : any
        Input object.

    Returns
    -------
    list or type of x
        If `x` can be iterated over, `x` will be returned as-is. Else,
        will return `[x]`, i.e. a list containing `x`.

    Examples
    --------
    >>> wrap_if_not_iterable(1.0)
    [1.0]
    >>> wrap_if_not_iterable('Hello, World!')
    'Hello, World!'
    """
    if is_iterable(x):
        return x
    else:
        return [x]


def nesting_level(x):
    """Calculate nesting level of a list of objects.

    To convert between sparse and dense representations of topological
    features, we need to determine the nesting level of an input list.
    The nesting level is defined as the maximum number of times we can
    recurse into the object while still obtaining lists.

    Parameters
    ----------
    x : list
        Input list of objects.

    Returns
    -------
    int
        Nesting level of `x`. If `x` has no well-defined nesting level,
        for example because `x` is not a list of something, will return
        `0`.

    Notes
    -----
    This function is implemented recursively. It is therefore a bad idea
    to apply it to objects with an extremely high nesting level.

    Examples
    --------
    >>> nesting_level([1, 2, 3])
    1

    >>> nesting_level([[1, 2], [3, 4]])
    2
    """
    # This is really only supposed to work with lists. Anything fancier,
    # for example a `torch.tensor`, can already be used as a dense data
    # structure.
    if not isinstance(x, list):
        return 0

    # Empty lists have a nesting level of 1.
    if len(x) == 0:
        return 1
    else:
        return max(nesting_level(y) for y in x) + 1



"""Example of classifying data using topological layers."""

from torch_topological.datasets import SphereVsTorus

from torch_topological.nn.data import make_tensor

from torch_topological.nn import VietorisRipsComplex
from torch_topological.nn.layers import StructureElementLayer

from torch.utils.data import DataLoader

from tqdm import tqdm

import torch


class TopologicalModel(torch.nn.Module):
    def __init__(self, n_elements, latent_dim=64, output_dim=2):
        super().__init__()

        self.n_elements = n_elements
        self.latent_dim = latent_dim

        self.model = torch.nn.Sequential(
            StructureElementLayer(self.n_elements),
            torch.nn.Linear(self.n_elements, self.latent_dim),
            torch.nn.ReLU(),
            torch.nn.Linear(self.latent_dim, output_dim),
        )

        self.vr = VietorisRipsComplex(dim=0)

    def forward(self, x):
        pers_info = self.vr(x)
        pers_info = make_tensor(pers_info)

        return self.model(pers_info)


if __name__ == "__main__":
    batch_size = 32
    n_epochs = 50
    n_elements = 10

    data_set = SphereVsTorus(n_point_clouds=2 * batch_size)

    loader = DataLoader(
        data_set,
        batch_size=batch_size,
        shuffle=True,
        drop_last=False,
    )

    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    model = TopologicalModel(n_elements).to(device)
    loss_fn = torch.nn.CrossEntropyLoss()
    opt = torch.optim.Adam(model.parameters(), lr=1e-4)

    progress = tqdm(range(n_epochs))

    for epoch in progress:
        for batch, (x, y) in enumerate(loader):
            x = x.to(device)
            y = y.to(device)

            output = model(x)
            loss = loss_fn(output, y)

            opt.zero_grad()
            loss.backward()
            opt.step()

            pred = torch.argmax(output, dim=1)
            acc = (pred == y).sum() / len(y)

            progress.set_postfix(loss=f"{loss.item():.08f}", acc=f"{acc:.02f}")



"""Layers and loss terms for persistence-based optimisation."""

from .data import PersistenceInformation

from .distances import WassersteinDistance
from .sliced_wasserstein_distance import SlicedWassersteinDistance
from .sliced_wasserstein_kernel import SlicedWassersteinKernel
from .multi_scale_kernel import MultiScaleKernel

from .loss import SignatureLoss
from .loss import SummaryStatisticLoss

from .alpha_complex import AlphaComplex
from .cubical_complex import CubicalComplex
from .vietoris_rips_complex import VietorisRipsComplex


__all__ = [
    'AlphaComplex',
    'CubicalComplex',
    'PersistenceInformation',
    'SignatureLoss',
    'SummaryStatisticLoss',
    'VietorisRipsComplex',
    'WassersteinDistance',
    'SlicedWassersteinDistance',
    'SlicedWassersteinKernel',
    'MultiScaleKernel',
]



"""Example demonstrating the computation of alpha complexes.

This simple example demonstrates how to use alpha complexes to change
the appearance of a point cloud, following the `TopologyLayer
<https://github.com/bruel-gabrielsson/TopologyLayer>`_ package.

This example is still a **work in progress**.
"""

from torch_topological.nn import AlphaComplex
from torch_topological.nn import SummaryStatisticLoss

from torch_topological.utils import SelectByDimension

import numpy as np
import matplotlib.pyplot as plt

import torch

if __name__ == '__main__':
    np.random.seed(42)
    data = np.random.rand(100, 2)

    alpha_complex = AlphaComplex()

    loss_fn = SummaryStatisticLoss(
        summary_statistic='polynomial_function',
        p=2,
        q=0
    )

    X = torch.nn.Parameter(torch.as_tensor(data), requires_grad=True)
    opt = torch.optim.Adam([X], lr=1e-2)

    for i in range(100):
        # We are only interested in working with persistence diagrams of
        # dimension 1.
        selector = SelectByDimension(1)

        # Let's think step by step; apparently, AIs like that! So let's
        # first get the persistence information of our complex. We pass
        # it through the selector to remove diagrams we do not need.
        pers_info = alpha_complex(X)
        pers_info = selector(pers_info)

        # Evaluate the loss; notice that we want to *maximise* it in
        # order to improve the holes in the data.
        loss = -loss_fn(pers_info)

        opt.zero_grad()
        loss.backward()
        opt.step()

    X = X.detach().numpy()

    plt.scatter(X[:, 0], X[:, 1])
    plt.show()



"""Demonstrate image smoothing based on topology."""

import numpy as np
import matplotlib.pyplot as plt

from torch_topological.nn import Cubical
from torch_topological.nn import SummaryStatisticLoss

from sklearn.datasets import make_circles

import torch


def _make_data(n_cells, n_samples=1000):
    X = make_circles(n_samples, shuffle=True, noise=0.05)[0]

    heatmap, *_ = np.histogram2d(X[:, 0], X[:, 1], bins=n_cells)
    heatmap -= heatmap.mean()
    heatmap /= heatmap.max()

    return heatmap


class TopologicalSimplification(torch.nn.Module):
    def __init__(self, theta):
        super().__init__()

        self.theta = theta

    def forward(self, x):
        persistence_information = cubical(x)
        persistence_information = [persistence_information[0]]

        gens, pd = persistence_information[0]

        persistence = (pd[:, 1] - pd[:, 0]).abs()
        indices = persistence <= self.theta

        gens = gens[indices]

        indices = torch.vstack((gens[:, 0:2], gens[:, 2:]))

        indices = np.ravel_multi_index(
            (indices[:, 0], indices[:, 1]), x.shape
        )

        x.ravel()[indices] = 0.0

        persistence_information = cubical(x)
        persistence_information = [persistence_information[0]]

        return x, persistence_information


if __name__ == '__main__':

    np.random.seed(23)

    Y = _make_data(50)
    Y = torch.as_tensor(Y, dtype=torch.float)
    X = torch.as_tensor(
        Y + np.random.normal(scale=0.05, size=Y.shape), dtype=torch.float
    )

    theta = torch.nn.Parameter(
        torch.as_tensor(1.0), requires_grad=True,
    )

    topological_simplification = TopologicalSimplification(theta)

    optimizer = torch.optim.Adam(
        [theta], lr=1e-2
    )
    loss_fn = SummaryStatisticLoss('total_persistence', p=1)

    cubical = Cubical()

    persistence_information_target = cubical(Y)
    persistence_information_target = [persistence_information_target[0]]

    for i in range(500):
        X, persistence_information = topological_simplification(X)

        optimizer.zero_grad()

        loss = loss_fn(
            persistence_information,
            persistence_information_target
        )

        print(loss.item(), theta.item())

        theta.backward()
        optimizer.step()

    X = X.detach().numpy()

    plt.imshow(X)
    plt.show()



"""Demo for distance minimisations of a point cloud.

Note
----
This demonstration is a work in progress. It is not fully documented and
tested yet.
"""

from tqdm import tqdm

from torch_topological.data import sample_from_disk

from torch_topological.nn import VietorisRipsComplex
from torch_topological.nn import WassersteinDistance

import torch
import torch.optim as optim


import matplotlib.pyplot as plt


if __name__ == '__main__':
    n = 100

    X = sample_from_disk(r=0.5, R=0.6, n=n)
    Y = sample_from_disk(r=0.9, R=1.0, n=n)

    X = torch.nn.Parameter(torch.as_tensor(X), requires_grad=True)

    vr = VietorisRipsComplex(dim=1)

    pi_target = vr(Y)
    loss_fn = WassersteinDistance(q=2)

    opt = optim.SGD([X], lr=0.1)

    n_iterations = 500
    progress = tqdm(range(n_iterations))

    for i in progress:

        opt.zero_grad()

        pi_source = vr(X)
        loss = loss_fn(pi_source, pi_target)

        loss.backward()
        opt.step()

        progress.set_postfix(loss=loss.item())

    X = X.detach().numpy()

    plt.scatter(X[:, 0], X[:, 1], label='Source')
    plt.scatter(Y[:, 0], Y[:, 1], label='Target')

    plt.legend()
    plt.show()



"""Sliced Wasserstein kernel implementation."""


import torch

from torch_topological.nn import SlicedWassersteinDistance

from torch_topological.utils import wrap_if_not_iterable


class SlicedWassersteinKernel(torch.nn.Module):
    """Calculate sliced Wasserstein kernel between persistence diagrams.

    This is an implementation of the sliced Wasserstein kernel between
    persistence diagrams, following [Carriere17a]_.

    References
    ----------
    .. [Carriere17a] M. Carrière et al., "Sliced Wasserstein Kernel for
       Persistence Diagrams", *Proceedings of the 34th International
       Conference on Machine Learning*, PMLR 70, pp. 664--673, 2017.
    """

    def __init__(self, num_directions=10, sigma=1.0):
        """Create new sliced Wasserstein kernel module.

        Parameters
        ----------
        num_directions : int
            Specifies the number of random directions to be sampled for
            computation of the sliced Wasserstein distance.

        sigma : int
            Variance term of the sliced Wasserstein kernel expression.
        """
        super().__init__()

        self.num_directions = num_directions
        self.sigma = sigma

    def forward(self, X, Y):
        """Calculate sliced Wasserstein kernel based on input tensors.

        Parameters
        ----------
        X : list or instance of :class:`PersistenceInformation`
            Topological features of the first space. Supposed to contain
            persistence diagrams and persistence pairings.

        Y : list or instance of :class:`PersistenceInformation`
            Topological features of the second space. Supposed to
            contain persistence diagrams and persistence pairings.

        Returns
        -------
        torch.tensor
            A single scalar tensor containing the sliced Wasserstein kernel
            between the persistence diagram(s) contained in `X` and `Y`.
        """
        total_cost = 0.0

        X = wrap_if_not_iterable(X)
        Y = wrap_if_not_iterable(Y)

        swd = SlicedWassersteinDistance(num_directions=self.num_directions)

        for pers_info in zip(X, Y):
            D1 = pers_info[0]
            D2 = pers_info[1]

            total_cost += torch.exp(-swd(D1, D2)) / self.sigma

        return total_cost



"""Contains multi-scale kernel (scale space) kernel module."""

import torch

from torch_topological.utils import wrap_if_not_iterable


class MultiScaleKernel(torch.nn.Module):
    # TODO: more detailed description
    r"""Implement the multi-scale kernel between two persistence diagrams.

    This class implements the multi-scale kernel between two persistence
    diagrams (also known as the scale space kernel) as defined by
    Reininghaus et al. [Reininghaus15a]_ as

    .. math::
        k_\sigma(F,G) = \frac{1}{8 \pi \sigma}
        \sum_{\substack{p \in F\\q \in G}} exp{-\frac{\|p-q\|^2}{8\sigma}}
        - exp{-\frac{\|p-\overline{q}\|^2}{8\sigma}}

    where :math:`z=(z_1, z_2)` and :math:`\overline{z}=(z_2, z_1)`

    References
    ----------
    .. [Reininghaus15a] J. Reininghaus, U. Bauer and R. Kwitt, "A Stable
       Multi-Scale Kernel for Topological Machine Learning", *Proceedings
       of the IEEE Conference on Computer Vision and Pattern Recognition*,
       pp. 4741--4748, 2015.
    """

    def __init__(self, sigma):
        """Create new  instance of the kernel.

        Parameters
        ----------
        sigma : float
            scale parameter of the kernel
        """
        super().__init__()

        self.sigma = sigma

    @staticmethod
    def _check_upper(d):
        # Check if all points in the diagram are above the diagonal.
        # All points below the diagonal are 'swapped'.
        is_upper = d[:, 0] < d[:, 1]
        if not torch.all(is_upper):
            d[~is_upper, 0] = d[~is_upper, 1]
            d[~is_upper, 1] = d[~is_upper, 0]
        return d

    @staticmethod
    def _mirror(x):
        # Mirror one or multiple points of a persistence
        # diagram at the diagonal
        if len(x.shape) > 1:
            return x[:, [1, 0]]
        # only a single point in the diagram
        return x[[1, 0]]

    @staticmethod
    def _dist(x, y, p):
        # Compute the point-wise lp-distance between two
        # persistence diagrams
        dist = torch.cdist(x, y, p=p)
        return dist.pow(2)

    def forward(self, X, Y, p=2.):
        """Calculate the kernel value between two persistence diagrams.

        The kernel value is computed for each dimension of the persistence
        diagram individually, according to Equation 10 from Reininghaus et al.
        The final kernel value is computed as the sum of kernel values over
        all dimensions.

        Parameters
        ----------
        X : list or instance of :class:`PersistenceInformation`
            Topological features of the first space. Supposed to
            contain persistence diagrams and persistence pairings.

        Y : list or instance of :class:`PersistenceInformation`
            Topological features of the second space. Supposed to
            contain persistence diagrams and persistence pairings.

        p : float or inf, default 2.
            Specify which p-norm to use for distance calculation.
            For infinity/maximum norm pass p=float('inf').
            Please note that using norms other than the 2-norm
            (Euclidean norm) are not guaranteed to give positive
            definite results.

        Returns
        -------
        torch.tensor
            A single scalar tensor containing the kernel value between the
            persistence diagram(s) contained in `X` and `Y`.

        Examples
        --------
        >>> from torch_topological.data.shapes import sample_from_disk
        >>> from torch_topological.nn import VietorisRipsComplex
        >>> # sample randomly from two disks
        >>> x = sample_from_disk(r=0.5, R=0.6, n=100)
        >>> y = sample_from_disk(r=0.9, R=1.0, n=100)
        >>> # compute vietoris rips filtration for both point clouds
        >>> vr = VietorisRipsComplex(dim=1)
        >>> vr_x = vr(x)
        >>> vr_y = vr(y)
        >>> # compute kernel value between persistence
        >>> # diagrams with sigma set to 1
        >>> msk = MultiScaleKernel(1.)
        >>> msk_value = msk(vr_x, vr_y)
        """
        X_ = wrap_if_not_iterable(X)
        Y_ = wrap_if_not_iterable(Y)

        k_sigma = 0.0

        for pers_info in zip(X_, Y_):
            # ensure that all points in the diagram are
            # above the diagonal
            D1 = self._check_upper(pers_info[0].diagram)
            D2 = self._check_upper(pers_info[1].diagram)

            # compute the pairwise distances between the
            # two diagrams
            nom = self._dist(D1, D2, p)
            # distance between diagram 1 and mirrored
            # diagram 2
            denom = self._dist(D1, self._mirror(D2), p)

            M = torch.exp(-nom) / (8 * self.sigma)
            M -= torch.exp(-denom) / (8 * self.sigma)

            # sum over all points
            k_sigma += M.sum() / (8. * self.sigma * torch.pi)

        return k_sigma



"""Loss terms for various optimisation objectives."""

import torch

from torch_topological.utils import is_iterable


class SummaryStatisticLoss(torch.nn.Module):
    r"""Implement loss based on summary statistic.

    This is a generic loss function based on topological summary
    statistics. It implements a loss of the following form:

    .. math:: \|s(X) - s(Y)\|^p

    In the preceding equation, `s` refers to a function that results in
    a scalar-valued summary of a persistence diagram.
    """

    def __init__(self, summary_statistic='total_persistence', **kwargs):
        """Create new loss function based on summary statistic.

        Parameters
        ----------
        summary_statistic : str
            Indicates which summary statistic function to use. Must be
            a summary statistics function that exists in the utilities
            module, i.e. :mod:`torch_topological.utils`.

            At present, the following choices are valid:

            - `torch_topological.utils.persistent_entropy`
            - `torch_topological.utils.polynomial_function`
            - `torch_topological.utils.total_persistence`
            - `torch_topological.utils.p_norm`

        **kwargs
            Optional keyword arguments, to be passed to the
            summary statistic function.
        """
        super().__init__()

        self.p = kwargs.get('p', 1.0)
        self.kwargs = kwargs

        import torch_topological.utils.summary_statistics as stat
        self.stat_fn = getattr(stat, summary_statistic, None)

    def forward(self, X, Y=None):
        r"""Calculate loss based on input tensor(s).

        Parameters
        ----------
        X : list of :class:`PersistenceInformation`
            Source information. Supposed to contain persistence diagrams
            and persistence pairings.

        Y : list of :class:`PersistenceInformation` or `None`
            Optional target information. If set, evaluates a difference
            in loss functions as shown in the introduction. If `None`,
            a simpler variant of the loss will be evaluated.

        Returns
        -------
        torch.tensor
            Loss based on the summary statistic selected by the client.
            Given a statistic :math:`s`, the function returns the
            following expression:

            .. math:: \|s(X) - s(Y)\|^p

            In case no target tensor `Y` has been provided, the latter part
            of the expression amounts to `0`.
        """
        stat_src = self._evaluate_stat_fn(X)

        if Y is not None:
            stat_target = self._evaluate_stat_fn(Y)
            return (stat_target - stat_src).abs().pow(self.p)
        else:
            return stat_src.abs().pow(self.p)

    def _evaluate_stat_fn(self, X):
        """Evaluate statistic function for a given tensor."""
        return torch.sum(
            torch.stack([
                self.stat_fn(pers_info.diagram, **self.kwargs)
                for pers_info in X
            ])
        )


class SignatureLoss(torch.nn.Module):
    """Implement topological signature loss.

    This module implements the topological signature loss first
    described in [Moor20a]_. In contrast to the original code provided
    by the authors, this module also provides extensions to
    higher-dimensional generators if desired.

    The module can be used in conjunction with any set of generators and
    persistence diagrams, i.e. with any set of persistence pairings and
    persistence diagrams. At the moment, it is restricted to calculating
    a Minkowski distances for the loss calculation.

    References
    ----------
    .. [Moor20a] M. Moor et al., "Topological Autoencoders",
        *Proceedings of the 37th International Conference on Machine
        Learning*, PMLR 119, pp. 7045--7054, 2020.
    """

    def __init__(self, p=2, normalise=True, dimensions=0):
        """Create new loss instance.

        Parameters
        ----------
        p : float
            Exponent for the `p`-norm calculation of distances.

        normalise : bool
            If set, normalises distances for each point cloud. This can
            be useful when working with batches.

        dimensions : int or tuple of int
            Dimensions to use in the signature calculation. Following
            [Moor20a]_, this is set by default to `0`.
        """
        super().__init__()

        self.p = p
        self.normalise = normalise
        self.dimensions = dimensions

        # Ensure that we can iterate over the dimensions later on, as
        # this simplifies the code.
        if not is_iterable(self.dimensions):
            self.dimensions = [self.dimensions]

    # TODO: Improve documentation and nomenclature.
    def forward(self, X, Y):
        """Calculate signature loss between two data sets."""
        X_pc, X_pi = X
        Y_pc, Y_pi = Y

        X_dist = torch.cdist(X_pc, X_pc, self.p)
        Y_dist = torch.cdist(Y_pc, Y_pc, self.p)

        if self.normalise:
            X_dist = X_dist / X_dist.max()
            Y_dist = Y_dist / Y_dist.max()

        X_sig_X = [
            self._select_distances(X_dist, X_pi[dim].pairing)
            for dim in self.dimensions
        ]
        X_sig_Y = [
            self._select_distances(X_dist, Y_pi[dim].pairing)
            for dim in self.dimensions
        ]
        Y_sig_X = [
            self._select_distances(Y_dist, X_pi[dim].pairing)
            for dim in self.dimensions
        ]
        Y_sig_Y = [
            self._select_distances(Y_dist, Y_pi[dim].pairing)
            for dim in self.dimensions
        ]

        XY_dist = self._partial_distance(X_sig_X, Y_sig_X)
        YX_dist = self._partial_distance(Y_sig_Y, X_sig_Y)

        return torch.stack(XY_dist).sum() + torch.stack(YX_dist).sum()

    def _select_distances(self, dist, gens):
        # Dimension 0: only a mapping of vertices--edges is present, and
        # we must *only* access the edges.
        if gens.shape[1] == 3:
            selected_distances = dist[gens[:, 1], gens[:, 2]]

        # Dimension > 0: we can access all distances
        else:
            creator_distances = dist[gens[:, 0], gens[:, 1]]
            destroyer_distances = dist[gens[:, 2], gens[:, 3]]

            # Need to use `torch.abs` here because of the way the
            # signature lookup works. We are *not* guaranteed  to
            # get 'valid' persistence values when using a pairing
            # from space X to access distances from space Y,  for
            # instance, hence some of values could be *negative*.
            selected_distances = torch.abs(
                destroyer_distances - creator_distances
            )

        return selected_distances

    def _partial_distance(self, A, B):
        """Calculate partial distances between pairings.

        The purpose of this function is to calculate a partial distance
        for the loss, depending on distances selected from the pairing.
        """
        dist = [
            0.5 * torch.linalg.vector_norm(a - b, ord=self.p)
            for a, b in zip(A, B)
        ]

        return dist



"""Filter functions for algorithmic outputs."""

import torch


class SelectByDimension(torch.nn.Module):
    """Select persistence diagrams by dimension.

    This is a simple selector that enables filtering the outputs of
    persistent homology calculation algorithms by dimension: often,
    one is really only interested in a *specific* dimension but the
    corresponding algorithm yields more diagrams. To this end, this
    module can be applied as a lightweight filter.
    """

    def __init__(self, min_dim, max_dim=None):
        """Prepare filter for subsequent usage.

        Provides the filter with the required parameters. A minimum
        dimension must be provided. There is also the option to use
        a maximum dimension, thus permitting filtering by ranges.

        Parameters
        ----------
        min_dim : int
            Minimum dimension to allow through the filter. If this is
            the sole provided parameter, only diagrams satisfying the
            dimension requirement will be selected.

        max_dim : int
            Optional upper dimension. If set, the selection returns a
            diagram whose dimension is within `[min_dim, max_dim]`.
        """
        super().__init__()

        self.min_dim = min_dim
        self.max_dim = max_dim

    def forward(self, X):
        """Apply selection parameters to input.

        Iterate over input and select diagrams according to the
        pre-defined parameters.

        Parameters
        ----------
        X : iterable of `PersistenceInformation`
            An iterable containing `PersistenceInformation` objects at
            its lowest level.

        Returns
        -------
        iterable
            Input, i.e. `X`, but with all non-matching persistence
            diagrams removed.
        """
        return [
            pers_info for pers_info in X if self._is_valid(pers_info)
        ]

    def _is_valid(self, pers_info):
        if self.max_dim is not None:
            return self.min_dim <= pers_info.dimension <= self.max_dim
        else:
            return self.min_dim == pers_info.dimension



"""Vietoris--Rips complex calculation module(s)."""

from itertools import starmap

from gph import ripser_parallel
from torch import nn

from torch_topological.nn import PersistenceInformation
from torch_topological.nn.data import batch_handler

import numpy
import torch


class VietorisRipsComplex(nn.Module):
    """Calculate Vietoris--Rips complex of a data set.

    This module calculates 'differentiable' persistence diagrams for
    point clouds. The underlying topological approximations are done
    by calculating a Vietoris--Rips complex of the data.
    """

    def __init__(
        self,
        dim=1,
        p=2,
        threshold=numpy.inf,
        keep_infinite_features=False,
        **kwargs
    ):
        """Initialise new module.

        Parameters
        ----------
        dim : int
            Calculates persistent homology up to (and including) the
            prescribed dimension.

        p : float
            Exponent indicating which Minkowski `p`-norm to use for the
            calculation of pairwise distances between points. Note that
            if `treat_as_distances` is supplied to :func:`forward`, the
            parameter is ignored and will have no effect. The rationale
            is to permit clients to use a pre-computed distance matrix,
            while always falling back to Minkowski norms.

        threshold : float
            If set to a finite number, only calculates topological
            features up to the specified distance threshold. Thus,
            any persistence pairings may contain infinite features
            as well.

        keep_infinite_features : bool
            If set, keeps infinite features. This flag is disabled by
            default. The rationale for this is that infinite features
            require more deliberate handling and, in case `threshold`
            is not changed, only a *single* infinite feature will not
            be considered in subsequent calculations.

        **kwargs
            Additional arguments to be provided to ``ripser``, i.e. the
            backend for calculating persistent homology. The `n_threads`
            parameter, which controls parallelisation, is probably the
            most relevant parameter to be adjusted.
            Please refer to the `the gitto-ph documentation
            <https://giotto-ai.github.io/giotto-ph/build/html/index.html>`_
            for more details on admissible parameters.

        Notes
        -----
        This module currently only supports Minkowski norms. It does not
        yet support other metrics internally. To use custom metrics, you
        need to set `treat_as_distances` in the :func:`forward` function
        instead.
        """
        super().__init__()

        self.dim = dim
        self.p = p
        self.threshold = threshold
        self.keep_infinite_features = keep_infinite_features

        # Ensures that the same parameters are used whenever calling
        # `ripser`.
        self.ripser_params = {
            'return_generators': True,
            'maxdim': self.dim,
            'thresh': self.threshold
        }

        self.ripser_params.update(kwargs)

    def forward(self, x, treat_as_distances=False):
        """Implement forward pass for persistence diagram calculation.

        The forward pass entails calculating persistent homology on
        a point cloud and returning a set of persistence diagrams.

        Parameters
        ----------
        x : array_like
            Input point cloud(s). `x` can either be a 2D array of shape
            `(n, d)`, which is treated as a single point cloud, or a 3D
            array/tensor of the form `(b, n, d)`, with `b` representing
            the batch size. Alternatively, you may also specify a list,
            possibly containing point clouds of non-uniform sizes.

        treat_as_distances : bool
            If set, treats `x` as containing pre-computed distances
            between points. The semantics of how `x` is handled are
            not changed; the only difference is that when `x` has a
            shape of `(n, d)`, the values of `n` and `d` need to be
            the same.

        Returns
        -------
        list of :class:`PersistenceInformation`
            List of :class:`PersistenceInformation`, containing both the
            persistence diagrams and the generators, i.e. the
            *pairings*, of a certain dimension of topological features.
            If `x` is a 3D array, returns a list of lists, in which the
            first dimension denotes the batch and the second dimension
            refers to the individual instances of
            :class:`PersistenceInformation` elements.

            Generators will be represented in the persistence pairing
            based on vertex--edge pairs (dimension 0) or edge--edge
            pairs. Thus, the persistence pairing in dimension zero will
            have three components, corresponding to a vertex and an
            edge, respectively, while the persistence pairing for higher
            dimensions will have four components.
        """
        return batch_handler(
            x,
            self._forward,
            treat_as_distances=treat_as_distances
        )

    def _forward(self, x, treat_as_distances=False):
        """Handle a *single* point cloud.

        This internal function handles the calculation of topological
        features for a single point cloud, i.e. an `array_like` of 2D
        shape.

        Parameters
        ----------
        x : array_like of shape `(n, d)`
            Single input point cloud.

        treat_as_distances : bool
            Flag indicating whether `x` should be treated as a distance
            matrix. See :func:`forward` for more information.

        Returns
        -------
        list of class:`PersistenceInformation`
            List of persistence information data structures, containing
            the persistence diagram and the persistence pairing of some
            dimension in the input data set.
        """
        if treat_as_distances:
            distances = x
        else:
            distances = torch.cdist(x, x, p=self.p)

        generators = ripser_parallel(
            distances.cpu().detach().numpy(),
            metric='precomputed',
            **self.ripser_params
        )['gens']

        # We always have 0D information.
        persistence_information = \
            self._extract_generators_and_diagrams(
                distances,
                generators,
                dim0=True,
            )

        if self.keep_infinite_features:
            persistence_information_inf = \
                self._extract_generators_and_diagrams(
                    distances,
                    generators,
                    finite=False,
                    dim0=True,
                )

        # Check whether we have any higher-dimensional information that
        # we should return.
        if self.dim >= 1:
            persistence_information.extend(
                self._extract_generators_and_diagrams(
                    distances,
                    generators,
                    dim0=False,
                )
            )

            if self.keep_infinite_features:
                persistence_information_inf.extend(
                    self._extract_generators_and_diagrams(
                        distances,
                        generators,
                        finite=False,
                        dim0=False,
                    )
                )

        # Concatenation is only necessary if we want to keep infinite
        # features.
        if self.keep_infinite_features:
            persistence_information = self._concatenate_features(
                persistence_information, persistence_information_inf
            )

        return persistence_information

    def _extract_generators_and_diagrams(
            self,
            dist,
            gens,
            finite=True,
            dim0=False
    ):
        """Extract generators and persistence diagrams from raw data.

        This convenience function translates between the output of
        `ripser_parallel` and the required output of this function.
        """
        index = 1 if not dim0 else 0

        # Perform index shift to find infinite features in the tensor.
        if not finite:
            index += 2

        gens = gens[index]

        if dim0:
            if finite:
                # In a Vietoris--Rips complex, all vertices are created at
                # time zero.
                creators = torch.zeros_like(
                    torch.as_tensor(gens)[:, 0],
                    device=dist.device
                )

                destroyers = dist[gens[:, 1], gens[:, 2]]
            else:
                creators = torch.zeros_like(
                    torch.as_tensor(gens)[:],
                    device=dist.device
                )

                destroyers = torch.full_like(
                    torch.as_tensor(gens)[:],
                    torch.inf,
                    dtype=torch.float,
                    device=dist.device
                )

                inf_pairs = numpy.full(
                    shape=(gens.shape[0], 2), fill_value=-1
                )
                gens = numpy.column_stack((gens, inf_pairs))

            persistence_diagram = torch.stack(
                (creators, destroyers), 1
            )

            return [PersistenceInformation(gens, persistence_diagram, 0)]
        else:
            result = []

            for index, gens_ in enumerate(gens):
                # Dimension zero is handled differently, so we need to
                # use an offset here. Note that this is not used as an
                # index into the `gens` array any more.
                dimension = index + 1

                if finite:
                    creators = dist[gens_[:, 0], gens_[:, 1]]
                    destroyers = dist[gens_[:, 2], gens_[:, 3]]

                    persistence_diagram = torch.stack(
                        (creators, destroyers), 1
                    )
                else:
                    creators = dist[gens_[:, 0], gens_[:, 1]]

                    destroyers = torch.full_like(
                        torch.as_tensor(gens_)[:, 0],
                        torch.inf,
                        dtype=torch.float,
                        device=dist.device
                    )

                    # Create special infinite pairs; we pretend that we
                    # are concatenating with unknown edges here.
                    inf_pairs = numpy.full(
                        shape=(gens_.shape[0], 2), fill_value=-1
                    )
                    gens_ = numpy.column_stack((gens_, inf_pairs))

                persistence_diagram = torch.stack(
                    (creators, destroyers), 1
                )

                result.append(
                    PersistenceInformation(
                        gens_,
                        persistence_diagram,
                        dimension)
                )

        return result

    def _concatenate_features(self, pers_info_finite, pers_info_infinite):
        """Concatenate finite and infinite features."""
        def _apply(fin, inf):
            assert fin.dimension == inf.dimension

            diagram = torch.concat((fin.diagram, inf.diagram))
            pairing = numpy.concatenate((fin.pairing, inf.pairing), axis=0)
            dimension = fin.dimension

            return PersistenceInformation(
                pairing=pairing,
                diagram=diagram,
                dimension=dimension
            )

        return list(starmap(_apply, zip(pers_info_finite, pers_info_infinite)))



"""Contains sampling routines for various simple geometric objects."""

import numpy as np
import torch

from .utils import embed


def sample_from_disk(n=100, r=0.9, R=1.0, seed=None):
    """Sample points from disk.

    Parameters
    ----------
    n : int
        Number of points to sample.

    r: float
        Minimum radius, i.e. the radius of the inner circle of a perfect
        sampling.

    R : float
        Maximum radius, i.e. the radius of the outer circle of a perfect
        sampling.

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    torch.tensor of shape `(n, 2)`
        Tensor containing the sampled coordinates.
    """
    assert r <= R, RuntimeError('r > R')

    rng = np.random.default_rng(seed)

    length = rng.uniform(r, R, size=n)
    angle = np.pi * rng.uniform(0, 2, size=n)

    x = np.sqrt(length) * np.cos(angle)
    y = np.sqrt(length) * np.sin(angle)

    X = np.vstack((x, y)).T
    return torch.as_tensor(X)


def sample_from_unit_cube(n, d=3, seed=None):
    """Sample points uniformly from unit cube in `d` dimensions.

    Parameters
    ----------
    n : int
        Number of points to sample

    d : int
        Number of dimensions.

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    torch.tensor of shape `(n, d)`
        Tensor containing the sampled coordinates.
    """
    rng = np.random.default_rng(seed)
    X = rng.uniform(size=(n, d))

    return torch.as_tensor(X)


def sample_from_sphere(n=100, d=2, r=1, noise=None, ambient=None, seed=None):
    """Sample `n` data points from a `d`-sphere in `d + 1` dimensions.

    Parameters
    -----------
    n : int
        Number of data points in shape.

    d : int
        Dimension of the sphere.

    r : float
        Radius of sphere.

    noise : float or None
        Optional noise factor. If set, data coordinates will be
        perturbed by a standard normal distribution, scaled by
        `noise`.

    ambient : int or None
        Embed the sphere into a space with ambient dimension equal to
        `ambient`. The sphere is randomly rotated into this
        high-dimensional space.

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    torch.tensor
        Tensor of sampled coordinates. If `ambient` is set, array will be
        of shape `(n, ambient)`. Else, array will be of shape `(n, d + 1)`.

    Notes
    -----
    This function was originally authored by Nathaniel Saul as part of
    the `tadasets` package. [tadasets]_

    References
    ----------
    .. [tadasets] https://github.com/scikit-tda/tadasets

    """
    rng = np.random.default_rng(seed)
    data = rng.standard_normal((n, d+1))

    # Normalize points to the sphere
    data = r * data / np.sqrt(np.sum(data**2, 1)[:, None])

    if noise:
        data += noise * rng.standard_normal(data.shape)

    if ambient is not None:
        assert ambient > d
        data = embed(data, ambient)

    return torch.as_tensor(data)


def sample_from_torus(n, d=3, r=1.0, R=2.0, seed=None):
    """Sample points uniformly from torus and embed it in `d` dimensions.

    Parameters
    ----------
    n : int
        Number of points to sample

    d : int
        Number of dimensions.

    r : float
        Radius of the 'tube' of the torus.

    R : float
        Radius of the torus, i.e. the distance from the centre of the
        'tube' to the centre of the torus.

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    torch.tensor of shape `(n, d)`
        Tensor of sampled coordinates.
    """
    rng = np.random.default_rng(seed)
    angles = []

    while len(angles) < n:
        x = rng.uniform(0, 2 * np.pi)
        y = rng.uniform(0, 1 / np.pi)

        f = (1.0 + (r/R) * np.cos(x)) / (2 * np.pi)

        if y < f:
            psi = rng.uniform(0, 2 * np.pi)
            angles.append((x, psi))

    X = []

    for theta, psi in angles:
        a = R + r * np.cos(theta)
        x = a * np.cos(psi)
        y = a * np.sin(psi)
        z = r * np.sin(theta)

        X.append((x, y, z))

    X = np.asarray(X)
    return torch.as_tensor(X)


def sample_from_annulus(n, r, R, seed=None):
    """Sample points from a 2D annulus.

    This function samples `N` points from an annulus with inner radius `r`
    and outer radius `R`.

    Parameters
    ----------
    n : int
        Number of points to sample

    r : float
        Inner radius of annulus

    R : float
        Outer radius of annulus

    seed : int, instance of `np.random.Generator`, or `None`
        Seed for the random number generator, or an instance of such
        a generator. If set to `None`, the default random number
        generator will be used.

    Returns
    -------
    torch.tensor of shape `(n, 2)`
        Tensor containing sampled coordinates.
    """
    if r >= R:
        raise RuntimeError(
            'Inner radius must be less than or equal to outer radius'
        )

    rng = np.random.default_rng(seed)
    thetas = rng.uniform(0, 2 * np.pi, n)

    # Need to sample based on squared radii to account for density
    # differences.
    radii = np.sqrt(rng.uniform(r ** 2, R ** 2, n))

    X = np.column_stack((radii * np.cos(thetas), radii * np.sin(thetas)))
    return torch.as_tensor(X)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['torch_topological',
 'torch_topological.data',
 'torch_topological.datasets',
 'torch_topological.examples',
 'torch_topological.nn',
 'torch_topological.utils']

package_data = \
{'': ['*']}

install_requires = \
['POT>=0.8.0,<0.9.0',
 'giotto-ph>=0.2.0,<0.3.0',
 'gudhi>=3.4.1,<4.0.0',
 'matplotlib>=3.5.0,<4.0.0',
 'torch>=1.10.1']

setup_kwargs = {
    'name': 'torch-topological',
    'version': '0.1.7',
    'description': 'A framework for topological machine learning based on `pytorch`.',
    'long_description': '<img src="https://raw.githubusercontent.com/aidos-lab/pytorch-topological/main/torch_topological.svg" height=128 alt="`pytorch-topological` icon" />\n\n# `pytorch-topological`: A topological machine learning framework for `pytorch`\n\n[![Documentation](https://readthedocs.org/projects/pytorch-topological/badge/?version=latest)](https://pytorch-topological.readthedocs.io/en/latest/?badge=latest) [![Maintainability](https://api.codeclimate.com/v1/badges/397f53d1968f01b86e74/maintainability)](https://codeclimate.com/github/aidos-lab/pytorch-topological/maintainability) ![GitHub contributors](https://img.shields.io/github/contributors/aidos-lab/pytorch-topological) ![PyPI - License](https://img.shields.io/pypi/l/torch_topological) ![PyPI](https://img.shields.io/pypi/v/torch_topological) [![Tests](https://github.com/aidos-lab/pytorch-topological/actions/workflows/run_tests.yaml/badge.svg)](https://github.com/aidos-lab/pytorch-topological/actions/workflows/run_tests.yaml)\n\n`pytorch-topological` (or `torch_topological`) is a topological machine\nlearning framework for [PyTorch](https://pytorch.org). It aims to\ncollect *loss terms* and *neural network layers* in order to simplify\nbuilding the next generation of topology-based machine learning tools.\n\n# Topological machine learning in a nutshell \n\n*Topological machine learning* refers to a new class of machine learning\nalgorithms that are able to make use of topological features in data\nsets. In contrast to methods based on a purely geometrical point of\nview, topological features are capable of focusing on *connectivity\naspects* of a data set. This provides an interesting fresh perspective\nthat can be used to create powerful hybrid algorithms, capable of\nyielding more insights into data.\n\nThis is an *emerging research field*, firmly rooted in computational\ntopology and topological data analysis. If you want to learn more about\nhow topology and geometry can work in tandem, here are a few resources\nto get you started:\n\n- Amézquita et al., [*The Shape of Things to Come: Topological Data Analysis and Biology,\n  from Molecules to Organisms*](https://doi.org/10.1002/dvdy.175), Developmental Dynamics\n  Volume 249, Issue 7, pp. 816--833, 2020.\n\n- Hensel et al., [*A Survey of Topological Machine Learning Methods*](https://www.frontiersin.org/articles/10.3389/frai.2021.681108/full),\n  Frontiers in Artificial Intelligence, 2021.\n\n# Installation and requirements\n\n`torch_topological` requires Python 3.9. More recent versions might work\nbut necessitate building some dependencies by yourself; Python 3.9\ncurrently offers the smoothest experience.\nIt is recommended to use the excellent [`poetry`](https://python-poetry.org) framework\nto install `torch_topological`:\n\n```\npoetry add torch-topological\n```\n\nAlternatively, use `pip` to install the package:\n\n```\npip install -U torch-topological\n```\n\n# Usage\n\n`torch_topological` is still a work in progress. You can [browse the documentation](https://pytorch-topological.readthedocs.io)\nor, if code reading is more your thing, dive directly into [some example\ncode](./torch_topological/examples).\n\nHere is a list of *other* projects that are using `torch_topological`:\n\n- [SHAPR](https://github.com/marrlab/SHAPR_torch), a method for for\n  predicting the 3D cell shape of individual cells based on 2D\n  microscopy images\n\nThis list is incomplete---you can help expanding it by using\n`torch_topological` in your own projects! :innocent:\n\n# Contributing\n\nCheck out the [contribution guidelines](CONTRIBUTING.md) or the [road\nmap](ROADMAP.md) of the project.\n\n# Acknowledgements\n\nOur software and research does not exist in a vacuum. `pytorch-topological` is standing\non the shoulders of proverbial giants. In particular, we want to thank the\nfollowing projects for constituting the technical backbone of the\nproject:\n\n| [`giotto-tda`](https://github.com/giotto-ai/giotto-tda)       | [`gudhi`](https://github.com/GUDHI/gudhi-devel)<br />       |\n|---------------------------------------------------------------|-------------------------------------------------------------|\n| <img src="logos/giotto.jpg" height=128 alt="`giotto` icon" /> | <img src="logos/gudhi.png" height=128 alt="`GUDHI` icon" /> |\n\nFurthermore, `pytorch-topological` draws inspiration from several\nprojects that provide a glimpse into the wonderful world of topological\nmachine learning:\n\n- [`difftda`](https://github.com/MathieuCarriere/difftda) by [Mathieu Carrière](https://github.com/MathieuCarriere)\n\n- [`Ripser`](https://github.com/Ripser/ripser) by [Ulrich Bauer](https://github.com/ubauer)\n\n- [`Teaspoon`](https://lizliz.github.io/teaspoon/) by [Elizabeth Munch](https://elizabethmunch.com/) and her team\n\n- [`TopologyLayer`](https://github.com/bruel-gabrielsson/TopologyLayer) by [Rickard Brüel Gabrielsson](https://github.com/bruel-gabrielsson)\n\n- [`topological-autoencoders`](https://github.com/BorgwardtLab/topological-autoencoders) by [Michael Moor](https://github.com/mi92), [Max Horn](https://github.com/ExpectationMax), and [Bastian Rieck](https://github.com/Pseudomanifold)\n\n- [`torchph`](https://github.com/c-hofer/torchph) by [Christoph Hofer](https://github.com/c-hofer) and [Roland Kwitt](https://github.com/rkwitt)\n\nFinally, `pytorch-topological` makes heavy use of [`POT`](https://pythonot.github.io), the Python Optimal Transport Library.\nWe are indebted to the many contributors of all these projects.\n',
    'author': 'Bastian Rieck',
    'author_email': 'bastian@rieck.me',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9',
}


setup(**setup_kwargs)



"""Sliced Wasserstein distance implementation."""


import torch

import numpy as np

from torch_topological.utils import wrap_if_not_iterable


class SlicedWassersteinDistance(torch.nn.Module):
    """Calculate sliced Wasserstein distance between persistence diagrams.

    This is an implementation of the sliced Wasserstein distance between
    persistence diagrams, following [Carriere17a]_.

    This module calculates the sliced Wasserstein distance between two
    persistence diagrams. It is an efficient variant of the Wasserstein
    distance, and it is commonly used in the Sliced Wasserstein Kernel.
    It computes the expected value of the Wasserstein distance when the
    persistence diagram is projected on a random line passing through
    the origin.
    """

    def __init__(self, num_directions=10):
        """Create new sliced Wasserstein distance calculation module.

        Parameters
        ----------
        num_directions : int
            Specifies the number of random directions to be sampled for
            computation of the sliced Wasserstein distance.
        """
        super().__init__()

        # Generates num_directions number of lines with slopes randomly sampled
        # between -pi/2 and pi/2.
        self.num_directions = num_directions
        thetas = torch.linspace(-np.pi/2, np.pi/2, steps=self.num_directions+1)
        thetas = thetas[:-1]
        self.lines = torch.vstack([torch.tensor([torch.cos(i), torch.sin(i)],
                                  dtype=torch.float32) for i in thetas])

    def _emd1d(self, X, Y):
        # Compute Wasserstein Distance between two 1d-distributions.
        X, ind = torch.sort(X, dim=0)
        Y, ind = torch.sort(Y, dim=0)
        return torch.sum(torch.abs(torch.sub(X, Y)))

    def _project_diagram(self, D1, L):
        # Project persistence diagram D1 onto a given line L.
        return torch.stack([torch.dot(x, L)/torch.dot(L, L) for x in D1])

    def forward(self, X, Y):
        """Calculate sliced Wasserstein metric based on input tensors.

        Parameters
        ----------
        X : list or instance of :class:`PersistenceInformation`
            Topological features of the first space. Supposed to contain
            persistence diagrams and persistence pairings.

        Y : list or instance of :class:`PersistenceInformation`
            Topological features of the second space. Supposed to
            contain persistence diagrams and persistence pairings.

        Returns
        -------
        torch.tensor
            A single scalar tensor containing the sliced Wasserstein distance
            between the persistence diagram(s) contained in `X` and `Y`.
        """
        total_cost = 0.0

        X = wrap_if_not_iterable(X)
        Y = wrap_if_not_iterable(Y)

        for pers_info in zip(X, Y):
            D1 = pers_info[0].diagram.float()
            D2 = pers_info[1].diagram.float()

            # Auxiliary array to project onto diagonal.
            diag = torch.tensor([0.5, 0.5], dtype=torch.float32)

            # Project both the diagrams onto the diagonals.
            D1_diag = torch.vstack([torch.sum(x) * diag for x in D1])
            D2_diag = torch.vstack([torch.sum(x) * diag for x in D2])

            cost = 0.0

            for line in self.lines:
                proj_d1 = self._project_diagram(D1, line)
                proj_d2 = self._project_diagram(D2, line)

                proj_diag_d1 = self._project_diagram(D1_diag, line)
                proj_diag_d2 = self._project_diagram(D2_diag, line)

                cost += self._emd1d(torch.cat([proj_d1, proj_diag_d2]),
                                    torch.cat([proj_d2, proj_diag_d1]))

            cost /= self.num_directions

            total_cost += cost

        return total_cost



"""Utility functions for data set generation."""

import numpy as np


def embed(data, ambient=50):
    """Embed `data` in `ambient` dimensions.

    Parameters
    ----------
    data : array_like
        Input data set; needs to have shape `(n, d)`, i.e. samples are
        in the rows, dimensions are in the columns.

    ambient : int
        Dimension of embedding space. Must be greater than
        dimensionality of data.

    Returns
    -------
    array_like
        Input array of shape `(n, D)`, with `D = ambient`.

    Notes
    -----
    This function was originally authored by Nathaniel Saul as part of
    the `tadasets` package. [tadasets]_

    References
    ----------
    .. [tadasets] https://github.com/scikit-tda/tadasets
    """
    n, d = data.shape
    assert ambient > d

    base = np.zeros((n, ambient))
    base[:, :d] = data

    # construct a rotation matrix of dimension `ambient`.
    random_rotation = np.random.random((ambient, ambient))
    q, r = np.linalg.qr(random_rotation)

    base = np.dot(base, q)
    return base



"""Demo for calculating cubical complexes.

This example demonstrates how to perform topological operations on
a structured array, such as a grey-scale image.
"""

import numpy as np
import matplotlib.pyplot as plt

from torch_topological.nn import CubicalComplex
from torch_topological.nn import WassersteinDistance

from tqdm import tqdm

import torch


def sample_circles(n_cells, n_samples=1000):
    """Sample two nested circles and bin them.

    Parameters
    ----------
    n_cells : int
        Number of cells for the 2D histogram, i.e. the 'resolution' of
        the histogram.

    n_samples : int
        Number of samples to use for creating the nested circles
        coordinates.

    Returns
    -------
    np.ndarray of shape ``(n_cells, n_cells)``
        Structured array containing intensity values for the data set.
    """
    from sklearn.datasets import make_circles
    X = make_circles(n_samples, shuffle=True, noise=0.01)[0]

    heatmap, *_ = np.histogram2d(X[:, 0], X[:, 1], bins=n_cells)
    heatmap -= heatmap.mean()
    heatmap /= heatmap.max()

    return heatmap


if __name__ == '__main__':

    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    print('Device', device)

    np.random.seed(23)

    Y = sample_circles(50)
    Y = torch.as_tensor(Y, dtype=torch.float)
    X = torch.as_tensor(
        Y + np.random.normal(scale=0.20, size=Y.shape),
        dtype=torch.float,
        device=device,
    )
    Y = Y.to(device)
    X = torch.nn.Parameter(X, requires_grad=True).to(device)

    source = X.clone()

    optimizer = torch.optim.Adam([X], lr=1e-3)
    loss_fn = WassersteinDistance(q=2)

    cubical_complex = CubicalComplex()

    persistence_information_target = cubical_complex(Y)
    persistence_information_target = persistence_information_target[0]

    n_iter = 500
    progress = tqdm(range(n_iter))

    for i in progress:
        persistence_information = cubical_complex(X)
        persistence_information = persistence_information[0]

        optimizer.zero_grad()

        loss = loss_fn(
            persistence_information,
            persistence_information_target
        )

        loss.backward()
        optimizer.step()

        progress.set_postfix(loss=loss.item())

    source = source.detach().numpy()
    target = Y.cpu().detach().numpy()
    result = X.cpu().detach().numpy()

    fig, ax = plt.subplots(ncols=3)

    ax[0].imshow(source)
    ax[0].set_title('Source')

    ax[1].imshow(target)
    ax[1].set_title('Target')

    ax[2].imshow(result)
    ax[2].set_title('Result')

    plt.show()
